#ifndef cxxtls_csv_header_included_p
#define cxxtls_csv_header_included_p

//
// Copyright 2014, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

///@file
///   This file defines namespace cxxtls::csv which defines various functions
///   associated with parsing and generating comma separated values files.

namespace cxxtls
{
namespace csv
{
    struct exception 
       ///  An exception that occurs while performing a Comma Separated Values
       ///  operation.
    {
        char const *what_;

        exception(char const *msg)
        :  what_(msg)
        {
        }

        char const *what() const { return what_; }
    };


    ///   functions and classes useful in manipulating CSV files.

    template<class value_type, int separator=','>
    struct asciiWhiteSpace
       /// a functor class that detects spaces -- given the current separator
       /// character as defined by the template parameter.
    {
        bool operator() (value_type const &t) const
        {
           if(t == separator)  
           {
               // handle tabs and other characters as separators 
               // they are not whitespace
               return false;
           }

           return    (t == ' ') 
                  || (t == '\t')
                  || (t == '\r')
                  || (t == '\n')
                  || (t == '\a')
                  ;
        }
    };

    template<class StringType, class IsWhiteSpace>
    StringType trim(StringType const &s, IsWhiteSpace const &pred)
        ///  return a trimmed copy of the input.  That is, remove leading and
        ///   trailing blanks.
    {
        StringType rv;
    
        typename StringType::const_iterator f = s.begin(), l = s.end();
    
        while ((f != l) && pred(*f))
        {
                *f++;
        }
    
        while ((f != l))
        {
                --l;
    
                if ( !pred(*l) )
                {
                        ++l;
                        break;
                }
    
                // *l is either space or tab, here
        }
    
    
        while (f != l)
                rv += *f++;
    
        return rv;
    }

    template<class StringType>
    inline StringType trim(StringType const &input)
        /// return a trimmed string assuming that ascii whitespace characters
        /// are the delimiters of interest.
    {
       return trim(input, asciiWhiteSpace<typename StringType::value_type>() );
    }

    template<class StringType>
    inline StringType trim(StringType const &input, int const &separator)
        /// return a trimmed string assuming that ascii whitespace characters
        /// are the delimiters of interest.
    {
       if(separator == '\t')
          trim(input, asciiWhiteSpace<typename StringType::value_type, '\t'>() );
       else
       if(separator > ' ' && separator <= '~'
         )
       {
           return trim(input, asciiWhiteSpace<typename StringType::value_type>() );
       }
       else
           throw exception("Invalid separator character");

       return std::string(); // no can get here big guy
    }

template<class StringType, class Container>
bool split(StringType const &s, Container &output, int separator)
    ///<  split a comma or tab separated values string into appropriate components and
    ///   and append them to output.  Output will be empties prior to the start of
    ///   parsing the string.
    ///
    ///   @note  This function understands sql style quoting for both " and '.
    ///   @note  The returned strings are trimmed.
    ///
    ///
    ///   @param[in]   s          The string to split up
    ///   @param[out]  output     The container to receive the split components
    ///   @param[in]   separator  The character that divides the string into components.
    ///
    ///   @return true on success.  If false is returned, check the size of
    ///   output for the number fields in output.
{
        typename StringType::const_iterator start = s.begin(), end = s.end();

        output.clear();

        StringType tmp;

        while (start != end)
        {
            switch (*start)
            {
                case '"':
                case '\'':
                {
                        typename StringType::value_type endChar = *start++;
                
                        while (start != end)
                        {
                                if (*start == endChar)
                                {
                                        typename StringType::const_iterator next = start; ++next;
                
                                        if (next == end)
                                        {
                                                ++start;
                                                break;
                
                                        }
                                        else
                                        if (*next == endChar)
                                        {
                                            ++start;
                                            ++start;
                                            tmp += endChar;
                                        }
                                        else
                                        {
                                                // next was not endChar and not end of string
                                                ++start;
                                                break;
                                        }
                                }
                                else
                                        tmp += *start++;
                        }
                }
                break;
                
                case ',':
                {
                    tmp = trim(tmp, separator);
                
                    output.push_back(tmp);
                
                    tmp.clear();
                
                    ++start;
                }
                break;
                
                default:
                    tmp += *start++;
                break;
                
            } // switch
        
        } // while(start != end)

        tmp = trim(tmp, separator);

        if (!tmp.empty())
                output.push_back(tmp);

        return true;
}

template<class StringType, class Container>
bool split(StringType const &s, Container &output)
    ///<  split a comma separated values string into appropriate components and
    ///   and append them to output.  Output will be empties prior to the start of
    ///   parsing the string.
    ///
    ///   @note  This function understands sql style quoting for both " and '.
    ///   @note  The returned strings are trimmed.
    ///
    ///
    ///   @param[in]   s          The string to split up
    ///   @param[out]  output     The container to receive the split components
    ///
    ///   @return true on success.  If false is returned, check the size of
    ///   output for the number fields in output.
{
    return split(s,output,',');
}


template<class StringType>
StringType decorate(StringType const &input, bool preferSingleQuotes=true)
    ///  convert a string into a value which is acceptable in an sql style
    ///   CSV file.
    ///
    ///   When the string has no commas, single, or double quotes, then it will
    ///   be returned as is.  If one of the offending characters is found, it will
    ///   encoded for SQL csv format -- which works like this:  In sql CSV
    ///   format, it is impossible to represent an empty string using quotes -- you
    ///   just use the comma separator to indicate nothing in the field -- or
    ///   leave it out if it is the last field on the line.  Quotes are quoted
    ///   by using 2 of the outer quotes -- either single or double quotes
    ///   based on the perferSingleQuotes option.
    ///
    ///       
{
    StringType rv;

    if (input.find_first_of(",'\"") >= input.size())
            return input; // nothing to do, just return the input

    typename StringType::value_type quote = preferSingleQuotes ? '\'' : '"';

    rv += quote;

    typename StringType::const_iterator f = input.begin(), l = input.end();

    while (f != l)
    {
        typename StringType::value_type c = *f++;

        if (c == quote)
        {
           rv += quote;
           rv += quote;
        }
        else
           rv += c;
    }

    rv += quote;
    return rv;
        
}

template<class Container, class StringType>
void format(Container const &input, StringType &output, typename StringType::value_type separator=',', bool preferSingleQuotes=true)
    ///<  Convert a vector strings into CSV format -- separated by the specifed
    ///   separator character, which defaults to comma.  The individual strings in the
    ///   output processed by decorateCSV() before appending.
{
        StringType &rv = output; // just an alias for the parameter

        rv.clear();

        typename Container::const_iterator f = input.begin(), l = input.end();

        while (f != l)
        {
            StringType const &cur = *f++;

            rv += decorate(cur, preferSingleQuotes);

            if (f != l)
                rv += separator;
        }

}

} // csv
} // cxxtls
#endif
