#ifndef CXX_fmtd_header_included_p
#define CXX_fmtd_header_included_p

#include <iostream>
#include <portable_io.h>
#include <cxxtls/classTraits.h>

namespace cxxtls
{

///@file
/// This file defines the symbol, @ref cxxtls::fmtd, that can be used to invoke snprintf on single
/// data items easily in a C++ program without declaring a local buffer to hold the 
/// results.


namespace fmtdalgo
  ///
  ///  fmtdalgo defines symbols used by the fmtd interface. It is only needed because
  ///  you can't specialize a template in a class body -- DOH!  STUPID C++
  ///
{
  // you can't specialize a template inside a class body, so this namespace is required 
  // to perform the specializations below.

  // helper --------------------------------------------------------------------
  inline void int2ascii(char *&scan, int d)
     /// Converts the specified integer into an ascii string
     /// stored at the specified location.
     ///
     /// Does not null terminate the string.  
     ///
     /// Moves the pointer to the first byte after the generated text.
  {
    char buffer[26];

    char *end = &buffer[25];

    *end = 0;

    char *ptr = end;

    while(d)
    {
      *(--ptr) = '0' + (d%10);
      d /= 10;
    }

    while(ptr != end)
      *scan++ = *ptr++;


  }

  // helper --------------------------------------------------------------------
  inline void setup(char *&scan, 
		    int    fieldWidth, 
		    int    precision, 
		    bool   lJust, 
		    bool   padding, 
		    char   pad
		   )
     ///
     /// Setup most of the parameters to the control string used in 
     /// invoking sprintf.  Does not append a nul
  {
       *scan++ = '%';

       if(lJust && fieldWidth != 0)
         *scan++ = '-';

       if(!lJust && pad == '0')
         *scan++ = '0';

       if(fieldWidth)
         int2ascii(scan, fieldWidth);

       if(precision)
       {
         *scan++ = '.';
         int2ascii(scan, precision);
       }
  }

  // formatter --------------------------------------------------------------------
  template<class T, int fieldWidth, int precision, int lJust, int padding, int pad, int base, int errchr>
  struct StringFormatter
     ///
     ///  This functor formats strings in the way that
     ///  should be done for class fmtd<>
     ///
     ///  Template parameter T is ignored.  This class defines
     ///  two kinds of strings that it supports, not all kinds imaginable.
     ///
  {

    size_t operator()(std::string const &v,
		      char              *buffer,
		      int                buffsize
		     )
    {
      return (*this)(v.c_str(), buffer, buffsize);
    }


    size_t operator() (char const *v,
		     char      *buffer, 
		     int        buffsize
		    )
      /// Convert the specified value into an ascii string and return the length.
      /// Do not write more than buffsize bytes (including the trailing null).
      ///
      /// Obeys the template parameters.
      ///
      /// Fills the output with errchr if the data won't fit.
      ///
    {
       char cs[100];

       char *scan = &cs[0];

       setup(scan,
             fieldWidth, 
             precision, 
             lJust, 
             padding, 
             pad
            );


       *scan++ = 's';

       *scan++ = 0;

       size_t bytesWritten = snprintf(buffer, buffsize, cs, v);

       size_t errorThreshold = buffsize;

       if(fieldWidth)
	 errorThreshold = fieldWidth;

       if( bytesWritten > errorThreshold)
       {
         std::fill_n(buffer, fieldWidth, errchr);
         buffer[fieldWidth] = 0;

	 bytesWritten = fieldWidth;
       }

       return bytesWritten;

    }

  };



  // -------------------------------------------------------------------------
  template<class T, int fieldWidth, int precision, int lJust, int padding, int pad, int base, int errchr>
  struct FloatFormatter
     ///
     ///  This functor formats floats, doubles, and long doubles in the way that
     ///  should be done for class fmtd<>
     ///
  {
    size_t operator() (T const   &v, 
		     char      *buffer, 
		     int        buffsize
		    )
      /// Convert the specified value into an ascii string and return the length.
      /// Do not write more than buffsize bytes (including the trailing null).
      ///
      /// Obeys the template parameters.
      ///
      /// Fills the output with errchr if the data won't fit.
      ///
    {
       char cs[100];

       char *scan = &cs[0];

       setup(scan,
             fieldWidth, 
             precision, 
             lJust, 
             padding, 
             pad
            );


       *scan++ = 'L';

       if(!lJust && precision)
	  *scan++ = 'f';
       else
	  *scan++ = 'g';

       *scan++ = 0;

       long double x = v;

       size_t bytesWritten = snprintf(buffer, buffsize, cs, x);

       size_t errorThreshold = buffsize;

       if(fieldWidth)
	 errorThreshold = fieldWidth;

       if( bytesWritten > errorThreshold)
       {
         std::fill_n(buffer, fieldWidth, errchr);
         buffer[fieldWidth] = 0;

	 bytesWritten = fieldWidth;
       }

       return bytesWritten;

    }

  };

  // -------------------------------------------------------------------------
  template<class T, int fieldWidth, int precision, int lJust, int padding, int pad, int base, int errchr>
  struct IntFormatter
     ///
     ///  This functor formats ints, shorts, char, unsigned, long, long long, etc. in 
     ///  the way that should be done for class fmtd<>
     ///
  {
    size_t operator() (T const  &v, 
		     char     *buffer, 
		     int       buffsize 
		    )
      /// Convert the specified value into an ascii string and return the length.
      /// Do not write more than buffsize bytes (including the trailing null).
      ///
      /// Obeys the template parameters.
      ///
      /// Fills the output with errchr if the data won't fit.
      ///
    {
       char cs[100];

       char *scan = &cs[0];

       setup(scan,
	     fieldWidth, 
	     precision, 
	     lJust, 
	     padding, 
	     pad
	    );


       *scan++ = 'l';
       *scan++ = 'l';

       if(base == 16)
	 *scan++ = 'x';
       else
       if(base == 8)
	 *scan++ = 'o';
       else
       if(isUnsignedType<T>::value)
	 *scan++ = 'u';
       else
         *scan++ = 'd';

       *scan++ = 0;

       long long x = v;

       size_t errorThreshold = buffsize;

       if(fieldWidth)
	 errorThreshold = fieldWidth;

       size_t rv;

       if( (rv = snprintf(buffer, buffsize, cs, x) ) > errorThreshold)
       {
         std::fill_n(buffer, fieldWidth, errchr);
         buffer[fieldWidth] = 0;
	 rv = fieldWidth;
       }

       return rv;
    }
  };

  // =========================================================================
  template<int N, 
	   class T, 
	   int fieldWidth, 
	   int precision, 
	   int lJust, 
	   int padding, 
	   int pad, 
	   int base, 
	   int errchr
	  > 
  struct algo;
    ///< A class which defines a member type, type, which
    ///  actually implements the formatting algorithm for fmtd<>
    ///  below.


  // =========================================================================
  template<class T, 
	   int fieldWidth, 
	   int precision, 
	   int lJust, 
	   int padding, 
	   int pad, 
	   int base, 
	   int errchr
	  > 
  struct algo<bctIntegralType,
	      T, 
	      fieldWidth, 
	      precision, 
	      lJust, 
	      padding, 
	      pad, 
	      base, 
	      errchr
	     >
      /// A specialization for the algo template that is
      /// creates the integer formatting algorithm used
      /// by fmtd<> below.
  { 
    typedef IntFormatter<T, 
			 fieldWidth, 
			 precision, 
			 lJust, 
			 padding, 
			 pad, 
			 base, 
			 errchr
			> 
                        type;   
  };

  // ==========================================================================
  template<class T, 
	   int   fieldWidth, 
	   int   precision, 
	   int   lJust, 
	   int   padding, 
	   int   pad, 
	   int   base, 
	   int   errchr
	  > 
  struct algo<bctFloatType, 
	      T, 
	      fieldWidth, 
	      precision, 
	      lJust, 
	      padding, 
	      pad, 
	      base, 
	      errchr
	     >    
      /// A specialization for the algo template that is
      /// creates the integer formatting algorithm used
      /// by fmtd<> below.
  { 
    typedef FloatFormatter<T, 
			   fieldWidth, 
			   precision, 
			   lJust, 
			   padding, 
			   pad, 
			   base, 
			   errchr
			  > type; 
  };

  // -------------------------------------------------------------------------
  template<
	   int   fieldWidth, 
	   int   precision, 
	   int   lJust, 
	   int   padding, 
	   int   pad, 
	   int   base, 
	   int   errchr
	  > 
  struct algo<bctClassType,   // parm named N
	      char const *,   // parm named T
	      fieldWidth, 
	      precision, 
	      lJust, 
	      padding, 
	      pad, 
	      base, 
	      errchr
	     >    
      /// A specialization for the algo template that is
      /// creates the integer formatting algorithm used
      /// by fmtd<> below.
  { 
    typedef StringFormatter<std::string,  // ignored but needed
			   fieldWidth, 
			   precision, 
			   lJust, 
			   padding, 
			   pad, 
			   base, 
			   errchr
			  > type; 
  };



};


template<int WIDTH=0, int PRECISION=0, int PAD=' ', int BASE=10, int ERRCHR='@'>
struct fmtd
  ///
  ///  A numeric to ascii string conversion class based on snprintf.  The constructor 
  ///  for this class constructs an object containing a char buffer holding the 
  ///  formatted resultant string.
  ///
  ///  Template class fmtd<> defines a family of string converters that can 
  ///  be used like this:
  ///
  ///  @code
  ///
  ///     cout << "some text" << fmtd<>(37) << endl;
  ///     cout << "some text" << fmtd<20,4,'0'>(-1e8) << endl;
  ///     cout << fmtd<0,0,0,16>(20) << endl;
  ///     cout << fmtd<0,0,0,8>(20) << endl;
  ///
  ///  @endcode
  ///
  ///  In the above cases, the number 37 is coverted to a string using sprintf and is
  ///  append to the output stream, right after :some text", leading to lines
  ///  containing:
  ///
  ///  @code
  ///
  ///    some text37
  ///    some text-00000100000000.0000
  ///    14
  ///    24
  /// 
  ///  @endcode
  ///
  ///  The fmtd template has the following template parameters:
  ///
  ///  @param WIDTH -- defaults to 0, specifies the maximum width of the resultant
  ///                  converted text.  
  ///                  But specifying 0 for width means that the output width is 
  ///                  variable, not fixed.  If negative, the field with is assumed
  ///                  be left justified -- right justified is the default.
  ///                  See the snprintf man page for width and precision.
  ///                  
  ///
  ///  @param PRECISION -- varies between integer and float.  See the snprintf
  ///                      man page.
  ///                      
  ///
  ///  @param PAD -- specifies whether we are using blanks or zeros to pad the
  ///                resultant output buffer.  
  ///
  ///  @param BASE --   either 10, 8, 16.  Ignored for floating point numbers
  ///
  ///  @param ERRCHR -- specifies the text to put in the output field if
  ///
  ///  Note also that the data type of the object being converted is not a part of
  ///  the template signature.  Instead, the constructor to the fmtd class is templated
  ///  and that is where the actual conversion algorithm appears.
  ///
  ///  Usage advice:
  ///
  ///  You can use a fmtd<> object as a variable:
  ///
  ///  @code
  ///
  ///    fmtd<20>   someVar( (unsigned int)(3) );
  ///
  ///  @endcode
  ///
  ///  But generally, this is ill-advised.  The whole point of this class it 
  ///  to use the fmtd<parms> constructor in output and other string conversion
  ///  expressions.
  /// 
  ///
{
  // implementation details

  // interface -----------------------------------------------------------------
  enum constants  { MIN_WIDTH   = 25,
		    BUFF_SAFETY = 10,
		    POS_WIDTH   = (WIDTH < 0) ? -WIDTH : WIDTH,
		    FIELD_WIDTH = (PRECISION > POS_WIDTH ? PRECISION : ( (POS_WIDTH == 0) ? 0 : POS_WIDTH ) ),
		    BUFF_SIZE   = FIELD_WIDTH + BUFF_SAFETY,
		    PADDING     = (POS_WIDTH != 0) || (PRECISION != 0),
		    LJUST       = (WIDTH< 1),
                  };

private:

  char buffer_[ BUFF_SIZE ];

  size_t size_;

public:

  // interface -----------------------------------------------------------------

  template<class T>
  fmtd(T const &v)
  {
    using namespace fmtdalgo;

    typename algo< basicClassType<T>::value, T, FIELD_WIDTH, PRECISION, LJUST, PADDING, PAD, BASE, ERRCHR>::type doit;
       // If the above line of code fails to compile it is most likely
       // related to the type, T, that you are attempting to use as a 
       // parameter to ftmd<>(T)! 
       // 
       // fmtd<> only supports the same data types that sprintf supports!
       //
       // Well, ok, it supports std::string too! 

    size_ = doit(v,buffer_,BUFF_SIZE);
  }

  // interface -----------------------------------------------------------------

  fmtd(char const *s)
  {
    using namespace fmtdalgo;

    typename algo< bctClassType, char const *, FIELD_WIDTH, PRECISION, LJUST, PADDING, PAD, BASE, ERRCHR>::type doit;

    size_ = doit(s,buffer_,BUFF_SIZE);
  }

  // interface -----------------------------------------------------------------

  fmtd(std::string const &s)
  {
    using namespace fmtdalgo;

    typename algo< bctClassType, char const *, FIELD_WIDTH, PRECISION, LJUST, PADDING, PAD, BASE, ERRCHR>::type doit;

    size_ = doit(s.c_str(),buffer_,BUFF_SIZE);
  }



  // interface -----------------------------------------------------------------

  operator char const *() const { return buffer_; }

  // interface -----------------------------------------------------------------

  size_t size() const { return size_; }


  // interface -----------------------------------------------------------------

  friend std::ostream &operator<< ( std::ostream &s, fmtd const &f) 
  { 
     s << f.buffer_; 
     return s;
  }


};

} // namespace cxxtls
#endif
